# encoding: BINARY
### original file from the ruby-mqtt gem
### located at https://github.com/njh/ruby-mqtt/blob/master/lib/mqtt/packet.rb
### Copyright (c) 2009-2013 Nicholas J Humfrey

# Copyright (c) 2016-2017 Pierre Goudet <p-goudet@ruby-dev.jp>
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# and Eclipse Distribution License v1.0 which accompany this distribution.
#
# The Eclipse Public License is available at
#    https://eclipse.org/org/documents/epl-v10.php.
# and the Eclipse Distribution License is available at
#   https://eclipse.org/org/documents/edl-v10.php.
#
# Contributors:
#    Pierre Goudet - initial committer

module PahoMqtt
  module Packet
    class Publish < PahoMqtt::Packet::Base
      # Duplicate delivery flag
      attr_accessor :duplicate

      # Retain flag
      attr_accessor :retain

      # Quality of Service level (0, 1, 2)
      attr_accessor :qos

      # The topic name to publish to
      attr_accessor :topic

      # The data to be published
      attr_accessor :payload

      # Default attribute values
      ATTR_DEFAULTS = {
        :topic   => nil,
        :payload => ''
      }

      # Create a new Publish packet
      def initialize(args={})
        super(ATTR_DEFAULTS.merge(args))
      end

      def duplicate
        @flags[3]
      end

      # Set the DUP flag (true/false)
      def duplicate=(arg)
        if arg.kind_of?(Integer)
          @flags[3] = (arg == 0x1)
        else
          @flags[3] = arg
        end
      end

      def retain
        @flags[0]
      end

      # Set the retain flag (true/false)
      def retain=(arg)
        if arg.kind_of?(Integer)
          @flags[0] = (arg == 0x1)
        else
          @flags[0] = arg
        end
      end

      def qos
        (@flags[1] ? 0x01 : 0x00) | (@flags[2] ? 0x02 : 0x00)
      end

      # Set the Quality of Service level (0/1/2)
      def qos=(arg)
        @qos = arg.to_i
        if @qos < 0 || @qos > 2
          raise PahoMqtt::PacketFormatException.new(
                  "Invalid QoS value: #{@qos}")
        else
          @flags[1] = (arg & 0x01 == 0x01)
          @flags[2] = (arg & 0x02 == 0x02)
        end
      end

      # Get serialisation of packet's body
      def encode_body
        body = ''
        if @topic.nil? || @topic.to_s.empty?
          raise PahoMqtt::PacketFormatException.new(
                  "Invalid topic name when serialising packet")
        end
        body += encode_string(@topic)
        body += encode_short(@id) unless qos == 0
        body += payload.to_s.dup.force_encoding('ASCII-8BIT')
        return body
      end

      # Parse the body (variable header and payload) of a Publish packet
      def parse_body(buffer)
        super(buffer)
        @topic   = shift_string(buffer)
        @id      = shift_short(buffer) unless qos == 0
        @payload = buffer
      end

      # Check that fixed header flags are valid for this packet type
      # @private
      def validate_flags
        if qos == 3
          raise PahoMqtt::PacketFormatException.new(
                  "Invalid packet: QoS value of 3 is not allowed")
        end
        if qos == 0 && duplicate
          raise PahoMqtt::PacketFormatException.new(
                  "Invalid packet: DUP cannot be set for QoS 0")
        end
      end

      # Returns a human readable string, summarising the properties of the packet
      def inspect
        "\#<#{self.class}: " +
          "d#{duplicate ? '1' : '0'}, " +
          "q#{qos}, " +
          "r#{retain ? '1' : '0'}, " +
          "m#{id}, " +
          "'#{topic}', " +
          "#{inspect_payload}>"
      end

      protected

      def inspect_payload
        str = payload.to_s
        if str.bytesize < 16 && str =~ /^[ -~]*$/
          "'#{str}'"
        else
          "... (#{str.bytesize} bytes)"
        end
      end
    end
  end
end
